/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "skki1_0.h"
#include <limits.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	REGPATH_KEYMAP			MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\Keymap")
#define	REGSUBKEY_JMAP			MYTEXT("j-map")
#define	REGSUBKEY_JABBREVMAP	MYTEXT("j-abbrev-map")
#define	NUM_COLUMN_LIST_KEYBIND		(3)
#define	NUM_BIND_PER_MYSPECIAL_KEY	NUM_COLUMN_LIST_KEYBIND

/*========================================================================
 *	private structures
 */
typedef struct {
	LPMYSTR			m_pszText ;
	UINT			m_uKeyCode ;
}	TVKKEYBIND ;

typedef struct {
	int			m_nCurSel ;
	int			m_nJMapBind ;
	int			m_nJAbbrevBind ;
	BYTE*		m_pJMap ;
	BYTE*		m_pAbbrevMap ;
}	TDlgSelectKeyBindArg ;

/*========================================================================
 *	private function prototypes 
 */
static	int					skkimeConfig_key2item				(BYTE*, UINT) ;
static	INT_PTR				skkimeConfig_dlgKeyBindOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR				skkimeConfig_dlgKeyBindOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR				skkimeConfig_dlgKeyBindOnCommand	(HWND, UINT, WPARAM, LPARAM) ;
static	BOOL				skkimeConfig_doSelectKeyBind		(HWND, int, int) ;
static	INT_PTR CALLBACK	skkimeConfig_dlgSelectKeyBind		(HWND, UINT, WPARAM, LPARAM) ;
static	int CALLBACK		skkimeConfig_compareKeyBind			(LPARAM, LPARAM, LPARAM) ;

static	BOOL				skkimeConfig_initKeymap		(LPCMYSTR, BYTE*, BYTE*) ;
static	BOOL				skkimeConfig_updateKeymap	(LPCMYSTR, BYTE*, BYTE*) ;

/*========================================================================
 *	private global variables
 */
/*
 * uȃ[hvƁuJi[hv Default ̃L[oChB
 */
static const int	riDefaultSkkJMapBind [SIZE_MYKEYMAP]	= {
#if 0
	NFUNC_SET_MARK_COMMAND,			NFUNC_BEGINNING_OF_LINE,		
	NFUNC_BACKWARD_CHAR,			NFUNC_PREFIX_CHAR,			
	NFUNC_DELETE_CHAR,				NFUNC_END_OF_LINE,			
	NFUNC_FORWARD_CHAR,				NFUNC_KEYBOARD_QUIT,			
	NFUNC_DELETE_BACKWARD_CHAR,		NFUNC_TRY_COMPLETION,		
	NFUNC_KANAINPUT_MODE_ON,		NFUNC_KILL_LINE,				
	NFUNC_REDRAW,					NFUNC_NEWLINE,				
	NFUNC_NEXT_LINE,				NFUNC_INVALID_CHAR,			
	NFUNC_PREVIOUS_LINE,			NFUNC_TOGGLE_JISX0201,		
	NFUNC_INVALID_CHAR,				NFUNC_INVALID_CHAR,			
	NFUNC_TRANSPOSE_CHARS,			NFUNC_INVALID_CHAR,			
	NFUNC_INVALID_CHAR,				NFUNC_KILL_REGION,			
	NFUNC_PREFIX_CHAR,				NFUNC_YANK,					
	NFUNC_INVALID_CHAR,				NFUNC_PREFIX_CHAR,			
	NFUNC_INVALID_CHAR,				NFUNC_INVALID_CHAR,			
	NFUNC_INVALID_CHAR,				NFUNC_INVALID_CHAR,			
	NFUNC_START_HENKAN,				NFUNC_SELF_INSERT,			
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,			
	NFUNC_DISPLAY_CODE_FOR_CHAR_AT_POINT,	NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_INSERT_COMMA,				NFUNC_SELF_INSERT,
	NFUNC_INSERT_PERIOD,			NFUNC_ABBREV_INPUT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SELF_INSERT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_TODAY,					NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_ZENKAKU_EIJI,				NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT_SUBR,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SET_HENKAN_POINT,
	NFUNC_PURGE_FROM_JISYO,			NFUNC_SET_HENKAN_POINT,
	NFUNC_SET_HENKAN_POINT,			NFUNC_SELF_INSERT,
	NFUNC_INPUT_BY_CODE_OR_MENU,	NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_INSERT_A,	
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_INSERT_E,
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_INSERT_I,
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_MODE_OFF,					NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_TOGGLE_KANA,
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_INSERT_U,
	NFUNC_KANA_INPUT,				NFUNC_KANA_INPUT,
	NFUNC_PREVIOUS_CANDIDATE,		NFUNC_KANA_INPUT,
	NFUNC_KANA_INPUT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_SELF_INSERT,
	NFUNC_SELF_INSERT,				NFUNC_DELETE_CHAR,
	/*	special key ̕B*/
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
#else
	NFUNC_INVALID_CHAR
#endif
} ;

/*
 * uABBREV[hv Default L[oChB
 */
static	const int	riDefaultAbbrevMapBind [SIZE_MYKEYMAP] = {
#if 0
	NFUNC_INVALID_CHAR,			NFUNC_BEGINNING_OF_LINE,
	NFUNC_BACKWARD_CHAR,			NFUNC_PREFIX_CHAR,
	NFUNC_DELETE_CHAR,			NFUNC_END_OF_LINE,
	NFUNC_FORWARD_CHAR,			NFUNC_KEYBOARD_QUIT,
	NFUNC_DELETE_BACKWARD_CHAR,	NFUNC_TRY_COMPLETION,
	NFUNC_KANAINPUT_MODE_ON,		NFUNC_KILL_LINE,
	NFUNC_REDRAW,				NFUNC_NEWLINE,
	NFUNC_NEXT_LINE,				NFUNC_INVALID_CHAR,
	NFUNC_PREVIOUS_LINE,			NFUNC_ZENKAKU_HENKAN,
	NFUNC_INVALID_CHAR,			NFUNC_INVALID_CHAR,
	NFUNC_TRANSPOSE_CHARS,		NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,			NFUNC_KILL_REGION,
	NFUNC_PREFIX_CHAR,			NFUNC_YANK,
	NFUNC_INVALID_CHAR,			NFUNC_PREFIX_CHAR,
	NFUNC_INVALID_CHAR,			NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,			NFUNC_INVALID_CHAR,
	NFUNC_START_HENKAN,			NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_ABBREV_COMMA,			NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_ABBREV_PERIOD,			NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SELF_INSERT_CHARACTER,	NFUNC_DELETE_CHAR,
	/*	special key ̕B*/
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,	NFUNC_INVALID_CHAR,
#else
	NFUNC_INVALID_CHAR
#endif
} ;

static LPCMYSTR	srszKeyFuncName [NUM_SELECTABLE_FUNCTIONS] = {
#if 0
	MYTEXT ("self-insert-char"),
	MYTEXT ("j-self-insert"),
	MYTEXT ("j-zenkaku-insert"),
	MYTEXT ("j-display-code-for-char-at-point"),
	MYTEXT ("j-set-henkan-point"),
	MYTEXT ("j-set-henkan-point-subr"),
	MYTEXT ("j-insert-a"),
	MYTEXT ("j-insert-e"),
	MYTEXT ("j-insert-i"),
	MYTEXT ("j-insert-o"),
	MYTEXT ("j-insert-u"),
	MYTEXT ("j-kana-input"),
	MYTEXT ("j-start-henkan"),
	MYTEXT ("j-insert-comma"),
	MYTEXT ("j-insert-period"),
	MYTEXT ("j-purge-from-jisyo"),
	MYTEXT ("j-input-by-code-or-menu"),
	MYTEXT ("j-mode-off"),
	MYTEXT ("j-toggle-kana"),
	MYTEXT ("j-previous-candidate"),
	MYTEXT ("j-kakutei"),
	MYTEXT ("j-abbrev-input"),
	MYTEXT ("j-abbrev-period"),
	MYTEXT ("j-abbrev-comma"),
	MYTEXT ("j-zenkaku-eiji"),
	MYTEXT ("j-zenkaku-henkan"), 
	MYTEXT ("j-today"),
	MYTEXT ("j-mode-off-and-self-insert"),
	MYTEXT ("j-kanainput-mode-on"),
	MYTEXT ("j-newline"),
	MYTEXT ("set-mark-command"),
	MYTEXT ("forward-char"),
	MYTEXT ("backward-char"),
	MYTEXT ("delete-char"),
	MYTEXT ("backward-delete-char"),
	MYTEXT ("j-try-completion"),
	MYTEXT ("end-of-line"),
	MYTEXT ("beginning-of-line"),
	MYTEXT ("kill-line"),
	MYTEXT ("yank"),
	MYTEXT ("kill-region"),
	MYTEXT ("kill-ring-save"),
	MYTEXT ("exchange-point-and-mark"),
	MYTEXT ("previous-line"),
	MYTEXT ("next-line"),
	MYTEXT ("transpose-chars"),
	MYTEXT ("redraw"),
	MYTEXT ("prefix-char"),
	MYTEXT ("delete-region"),
	MYTEXT ("j-sendback-key"),
	MYTEXT ("keyboard-quit"),
	MYTEXT ("j-toggle-jisx0201"),
	MYTEXT ("not-modified"),
	MYTEXT ("forward-word"),
	MYTEXT ("backward-word"),
	MYTEXT ("upase-word"),
	MYTEXT ("downcase-word"),
	MYTEXT ("capitalize-word"),
	MYTEXT ("toggle-ime"),
	MYTEXT ("left-arrow"),
	MYTEXT ("up-arrow"),
	MYTEXT ("right-arrow"),
	MYTEXT ("down-arrow"),
	MYTEXT ("save-user-jisyo"),
#else
	MYTEXT ("invalid-char"),
#endif
} ;

static	const TVKKEYBIND	srVkKeyBinds [NUM_MYSPECIAL_KEY]	= {
	{ MYTEXT ("left"),		VK_LEFT, },
	{ MYTEXT ("up"),		VK_UP, },
	{ MYTEXT ("down"),		VK_DOWN, },
	{ MYTEXT ("right"),		VK_RIGHT, },
	{ MYTEXT ("insert"),	VK_INSERT, },
	{ MYTEXT ("delete"),	VK_DELETE, },
	{ MYTEXT ("backspace"),	VK_BACK, },
	{ MYTEXT ("tab"),		VK_TAB, },
	{ MYTEXT ("home"),		VK_HOME, },
	{ MYTEXT ("space"),		VK_SPACE, },
	{ MYTEXT ("escape"),	VK_ESCAPE, },
	{ MYTEXT ("kanji"),		VK_KANJI, },
	{ MYTEXT ("help"),		VK_HELP, },
	{ MYTEXT ("f1"),		VK_F1, },
	{ MYTEXT ("f2"),		VK_F2, },
	{ MYTEXT ("f3"),		VK_F3, },
	{ MYTEXT ("f4"),		VK_F4, },
	{ MYTEXT ("f5"),		VK_F5, },
	{ MYTEXT ("f6"),		VK_F6, },
	{ MYTEXT ("f7"),		VK_F7, },
	{ MYTEXT ("f8"),		VK_F8, },
	{ MYTEXT ("f9"),		VK_F9, },
	{ MYTEXT ("f10"),		VK_F10, },
	{ MYTEXT ("f11"),		VK_F11, },
	{ MYTEXT ("f12"),		VK_F12, },
	{ MYTEXT ("clear"),		VK_CLEAR, },
	{ MYTEXT ("return"),	VK_RETURN, },
} ;

/*========================================================================
 *	public functions
 */
BOOL
skkimeConfig_InitializeKeymap (
	register TINYMEMMGR*	pMemMgr,
	register BYTE*			pJMap,
	register BYTE*			pAbbrevMap)
{
	if (pJMap == NULL || pAbbrevMap == NULL)
		return	FALSE ;
	return	skkimeConfig_initKeymap (REGPATH_KEYMAP, pJMap, pAbbrevMap) ;
	UNREFERENCED_PARAMETER (pMemMgr) ;
}

BOOL
skkimeConfig_UpdateKeymap (
	register BYTE*			pJMap,
	register BYTE*			pAbbrevMap)
{
	if (pJMap == NULL || pAbbrevMap == NULL)
		return	FALSE ;
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateKeymap (%p, %p)\n"), pJMap, pAbbrevMap)) ;
	return	skkimeConfig_updateKeymap (REGPATH_KEYMAP, pJMap, pAbbrevMap) ;
}

/*	L[̔菇́A
 *	(1) zL[R[hւ bind `FbNB
 *	(2) R[hւ bind `FbNB
 *	łB
 *	āȀ͎ۂ keymap ЂɎĝłāAł͕svȂ
 *	ǂˁB
 */
void
skkimeConfig_key2string (
	register LPMYSTR	pstrBuffer,
	register UINT		uKey)
{
	static LPCMYSTR	rpFormat []	= {
		MYTEXT ("[%s]"), MYTEXT ("[C-%s]"), MYTEXT ("[S-%s]"),
	} ;

	if (/*0 <= uKey &&*/ uKey < 0x20) {
		Mywsprintf (pstrBuffer, MYTEXT ("\\C-%c"), MYTEXT ("@abcdefghijklmnopqrstuvwxyz[\\]^_") [uKey]) ;
		return ;
	}
	if (uKey == 0x20) {
		Mylstrcpy (pstrBuffer, MYTEXT ("\\40")) ;
		return ;
	}
	if (0x20 < uKey && uKey < 0x7F) {
		Mywsprintf (pstrBuffer, MYTEXT ("%c"), uKey) ;
		return ;
	}
	if (uKey == 0x7F) {
		Mylstrcpy (pstrBuffer, MYTEXT ("\\177")) ;
		return ;
	}
	if (0x80 <= uKey && uKey < SIZE_MYKEYMAP) {
		register int	nMajor, nMinor ;
		nMajor	= (uKey - 0x80) / NUM_BIND_PER_MYSPECIAL_KEY ;
		nMinor	= (uKey - 0x80) % NUM_BIND_PER_MYSPECIAL_KEY ;
		Mywsprintf (pstrBuffer, rpFormat [nMinor], srVkKeyBinds [nMajor].m_pszText) ;
		return ;
	}
	Mylstrcpy (pstrBuffer, MYTEXT ("undefined")) ;
	return ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgKeybindProc (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgKeybindProc (%x, %d, %x, %x)\n"),
				  hDlg, message, wParam, lParam)) ;

	switch (message) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgKeyBindOnInitDialog (hDlg, message, wParam, lParam) ;

	case	WM_NOTIFY:
		return	skkimeConfig_dlgKeyBindOnNotify (hDlg, message, wParam, lParam) ;

	case	WM_COMMAND:
		return	skkimeConfig_dlgKeyBindOnCommand (hDlg, message, wParam, lParam) ;

	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================
 *	Private functions
 */
INT_PTR
skkimeConfig_dlgKeyBindOnInitDialog (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register HWND				hDlgKeybind ;
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	static struct {
		LPMYSTR	m_strTitle ;
		int		m_nWidth ;
	}	srColumnInfo [NUM_COLUMN_LIST_KEYBIND]	= {
		{ MYTEXT ("key"),		100, },
		{ MYTEXT ("j-map"),		130, },
		{ MYTEXT ("j-abbrev"),	130, },
	} ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
	pPropSheet	= (LPPROPSHEETPAGE) lParam ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgKeyBindOnInitDialog: lParam(%p), pConfArg(%p)\n"), pPropSheet, pConfArg)) ;

	hDlgKeybind	= GetDlgItem (hDlg, IDC_LIST_KEYBIND) ;
	if (hDlgKeybind != NULL) {
		register int	nIndex, nKey ;
		MYLV_COLUMN		lvColumn;
		MYLVITEM			lvI ;

		MyListView_DeleteAllItems (hDlgKeybind) ;
		MyListView_SetExtendedListViewStyle (hDlgKeybind, LVS_EX_FULLROWSELECT) ;
		
		lvColumn.mask	= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
		lvColumn.fmt	= LVCFMT_LEFT ;
		
		// Insert the five columns in the list view.
		for (nIndex = 0; nIndex < NUM_COLUMN_LIST_KEYBIND ; nIndex ++) {
			lvColumn.cx			= srColumnInfo [nIndex].m_nWidth ;
			lvColumn.pszText	= srColumnInfo [nIndex].m_strTitle ;
			MyListView_InsertColumn (hDlgKeybind, nIndex, &lvColumn) ;
		}
		
		lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
		lvI.state		= 0 ; 
		lvI.stateMask	= 0 ;
		//nItem			= 0 ;
		
		// Initialize MYLVITEM members that are different for each item. 
		nIndex	= 0 ;
		for (nKey = 0 ; nKey < SIZE_MYKEYMAP ; nKey ++) {
			register int	nJMapBind, nAbbrevBind, nItem ;
			MYCHAR	szBuffer [32] ;
			
			nJMapBind	= skkimeConfig_key2item (pConfArg->m_rJMap,      nKey) ;
			nAbbrevBind	= skkimeConfig_key2item (pConfArg->m_rAbbrevMap, nKey) ;
			if ((0 <= nJMapBind && nJMapBind < NUM_SELECTABLE_FUNCTIONS &&
				 nJMapBind   != NFUNC_INVALID_CHAR &&
				 nJMapBind   != NFUNC_NOFUNCTION) ||
				(0 <= nAbbrevBind && nAbbrevBind < NUM_SELECTABLE_FUNCTIONS && 
				 nAbbrevBind != NFUNC_INVALID_CHAR &&
				 nAbbrevBind != NFUNC_NOFUNCTION)) {

				DEBUGPRINTF ((MYTEXT ("Key(%d), j-map(%d), abbrev-map(%d)\n"), 
							  nKey, nJMapBind, nAbbrevBind)) ;

				lvI.iItem		= nIndex ;
				lvI.iImage		= 0 ;
				lvI.iSubItem	= 0 ;
				lvI.lParam		= (LPARAM) nKey ;
				skkimeConfig_key2string (szBuffer, nKey) ;
				lvI.pszText		= szBuffer ;
				nItem			= MyListView_InsertItem (hDlgKeybind, &lvI) ;
				if (nItem != -1) {
					DEBUGPRINTF ((MYTEXT ("Item(%d) <- \"%s\", \"%s\"\n"),
								  nItem,
								  srszKeyFuncName [nJMapBind],
								  srszKeyFuncName [nAbbrevBind])) ;
					MyListView_SetItemText (hDlgKeybind, nItem, 1, (LPMYSTR) srszKeyFuncName [nJMapBind]) ;
					MyListView_SetItemText (hDlgKeybind, nItem, 2, (LPMYSTR) srszKeyFuncName [nAbbrevBind]) ;
				}
				nIndex	++ ;
			}
		}
	}
	return	FALSE ;

	UNREFERENCED_PARAMETER (message) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgKeyBindOnNotify (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;
	static TMYMENUITEM		rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0, TEXT("ǉ"), IDC_BUTTON_INSERT, },
		{ MIIM_ID | MIIM_STRING, 0, TEXT("ύX"), IDC_BUTTON_EDIT, },
		{ MIIM_ID | MIIM_STRING, 0, TEXT("폜"), IDC_BUTTON_DELETE, },
	} ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	switch (pNMHDR->idFrom) {
	case	IDC_LIST_KEYBIND:
	{
		if (pNMHDR->code == NM_RCLICK) {
			register NMITEMACTIVATE*	pNM 	= (NMITEMACTIVATE*) lParam ;
			register int	n ;

			n	= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
			DEBUGPRINTF ((MYTEXT ("NM: item(%d), lParam(%d)\n"), pNM->iItem, pNM->lParam)) ;
			if (n > 0) 
				skkimeConfig_doSelectKeyBind (hDlg, pNM->iItem, n) ;
			DEBUGPRINTF ((MYTEXT ("TrackPopupMenu returns %d\n"), n)) ;
		}
		return	FALSE ;
	}
	default:
		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateKeymap (pConfArg->m_rJMap, pConfArg->m_rAbbrevMap)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;
		default:
			return FALSE ;
		}
		return	TRUE ;
	}

	UNREFERENCED_PARAMETER (message) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgKeyBindOnCommand (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;

	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	DEBUGPRINTF ((MYTEXT ("HWND:(%lx), ID(%u), Notify(%u), CTLWND:(%lx)\n"),
				  hDlg, wID, wNotifyCode, hwndCtl)) ;

	if ((wID == IDC_BUTTON_INSERT || wID == IDC_BUTTON_EDIT || wID == IDC_BUTTON_DELETE) &&
		wNotifyCode == BN_CLICKED) {
		register HWND	hwndLV	= GetDlgItem (hDlg, IDC_LIST_KEYBIND) ;
		register int	nMark	= -1 ;
		if (hwndLV == NULL)
			return	1 ;
		if (wID != IDC_BUTTON_INSERT) {
			nMark	= MyListView_GetSelectionMark (hwndLV) ;
			if (nMark < 0)
				return	1 ;
		}
		skkimeConfig_doSelectKeyBind (hDlg, nMark, wID) ;
	}
	return 0 ;

	UNREFERENCED_PARAMETER (message) ;
}

/*	X̃L[oCh肷_CAOB
 */
BOOL
skkimeConfig_doSelectKeyBind (
	register HWND	hDlg,
	register int	nItem,
	register int	nControl)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register HWND				hwndLV ;
	register int				nRetval, nCurSel ;
	MYLVFINDINFO				lvfi ;
	MYLVITEM					lvI ;
	MYCHAR						szBuffer [32] ;

	pPropSheet		= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg		= (SkkimeConfigArg *) pPropSheet->lParam ;
	hwndLV			= GetDlgItem (hDlg, IDC_LIST_KEYBIND) ;
	if (hwndLV == NULL || pConfArg == NULL)
		return	FALSE ;
	lvI.mask		= LVIF_PARAM ;
	lvI.state		= 0 ;
	lvI.stateMask	= 0 ;
	lvI.iItem		= nItem ;
	lvI.iSubItem	= 0 ;
	lvI.pszText		= 0 ;
	lvI.cchTextMax	= 0 ;
	if (! MyListView_GetItem (hwndLV, &lvI)) {
		nCurSel		= -1 ;
	} else {
		nCurSel		= lvI.lParam ;
	}

	switch (nControl) {
	case	IDC_BUTTON_EDIT:
		if (nCurSel < 0 || nCurSel >= SIZE_MYKEYMAP)
			return	FALSE ;
		/*	fall through */

	case	IDC_BUTTON_INSERT:
	{
		TDlgSelectKeyBindArg	arg ;

		arg.m_nCurSel		= nCurSel ;
		arg.m_nJMapBind		= NFUNC_NOFUNCTION ;
		arg.m_nJAbbrevBind	= NFUNC_NOFUNCTION ;
		arg.m_pJMap			= pConfArg->m_rJMap ;
		arg.m_pAbbrevMap	= pConfArg->m_rAbbrevMap ;

		nRetval	= DialogBoxParam (pConfArg->m_hInst, MAKEINTRESOURCE (IDD_DIALOG_SELECTKEYBIND), hDlg, skkimeConfig_dlgSelectKeyBind, (LPARAM) &arg) ;
		if (nRetval == IDOK) {
			register int	nAbbrevBind, nJMapBind ;

			nCurSel		= arg.m_nCurSel ;
			nJMapBind	= arg.m_nJMapBind ;
			nAbbrevBind	= arg.m_nJAbbrevBind ;
			if ((0 <= nCurSel     && nCurSel     < SIZE_MYKEYMAP) &&
				(0 <= nJMapBind   && nJMapBind   < NUM_SELECTABLE_FUNCTIONS) &&
				(0 <= nAbbrevBind && nAbbrevBind < NUM_SELECTABLE_FUNCTIONS)) {
				/*	CurSel ω邱Ƃ̂ŁAnItem ͎蒼B*/
				memset (&lvfi, 0, sizeof (lvfi)) ;
				lvfi.flags		= LVFI_PARAM ;
				lvfi.lParam		= nCurSel ;
				nItem			= MyListView_FindItem (hwndLV, -1, &lvfi) ;

				lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
				lvI.state		= 0 ; 
				lvI.stateMask	= 0 ;
				lvI.iImage		= 0 ;
				lvI.iSubItem	= 0 ;
				lvI.lParam		= (LPARAM) nCurSel ;
				skkimeConfig_key2string (szBuffer, nCurSel) ;
				lvI.pszText		= szBuffer ;

				if (nItem < 0) {
					nItem			= MyListView_GetItemCount (hwndLV) ;
					lvI.iItem		= nItem ;
					MyListView_InsertItem (hwndLV, &lvI) ;
					MyListView_SetItemText (hwndLV, nItem, 1, (LPMYSTR) srszKeyFuncName [nJMapBind]) ;
					MyListView_SetItemText (hwndLV, nItem, 2, (LPMYSTR) srszKeyFuncName [nAbbrevBind]) ;
					MyListView_SortItems   (hwndLV, skkimeConfig_compareKeyBind, 0) ;
				} else {
					lvI.iItem		= nItem ;
					MyListView_SetItem     (hwndLV, &lvI) ;
					MyListView_SetItemText (hwndLV, nItem, 1, (LPMYSTR) srszKeyFuncName [nJMapBind]) ;
					MyListView_SetItemText (hwndLV, nItem, 2, (LPMYSTR) srszKeyFuncName [nAbbrevBind]) ;
				}
				pConfArg->m_rJMap		[nCurSel]	= (unsigned char)nJMapBind ;
				pConfArg->m_rAbbrevMap	[nCurSel]	= (unsigned char)nAbbrevBind ;
				MyListView_Update (hwndLV, nItem) ;
			}
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		}
		break ;
	}
	case	IDC_BUTTON_DELETE:
		if (0 <= nCurSel && nCurSel < SIZE_MYKEYMAP) {
			pConfArg->m_rJMap      [nCurSel]	= NFUNC_INVALID_CHAR ;
			pConfArg->m_rAbbrevMap [nCurSel]	= NFUNC_INVALID_CHAR ;
			if (! MyListView_DeleteItem (hwndLV, nItem)) {
				DEBUGPRINTF ((MYTEXT ("MyListView_DeleteItem (%lx, %d) failed\n"), 
							  hwndLV, nItem)) ;
			}
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		}
		break ;
	default:
		return	FALSE ;
	}
	return	TRUE ;
}

INT_PTR CALLBACK
skkimeConfig_dlgSelectKeyBind (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register HWND	hwndKeyToBind, hwndJMap, hwndJAbbrevMap ;
	register TDlgSelectKeyBindArg*	pArg ;
	MYCHAR							szBuffer [32] ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		register int	i, nCurSel, nJMapFunc, nJAbbrevFunc ;

		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;

		pArg			= (TDlgSelectKeyBindArg*) lParam ;
		if (pArg != NULL && 0 <= pArg->m_nCurSel && pArg->m_nCurSel < SIZE_MYKEYMAP) {
			nCurSel			= pArg->m_nCurSel ;
			nJMapFunc		= pArg->m_pJMap      [nCurSel] ;
			nJAbbrevFunc	= pArg->m_pAbbrevMap [nCurSel] ;
		} else {
			nCurSel			= -1 ;
			nJMapFunc		= -1 ;
			nJAbbrevFunc	= -1 ;
		}
		hwndKeyToBind	= GetDlgItem (hDlg, IDC_COMBO_KEYTOBIND) ;
		DEBUGPRINTF ((MYTEXT ("hwndKeyToBind (%lx), Select(%d), Map(%d/%d)\n"),
					  hwndKeyToBind, nCurSel, nJMapFunc, nJAbbrevFunc)) ;
		if (hwndKeyToBind != NULL) {
			/*	̗̈́uvƂĈB*/
			for (i = 0 ; i < SIZE_MYKEYMAP ; i ++) {
				skkimeConfig_key2string (szBuffer, i) ;
				SendMessage (hwndKeyToBind, CB_ADDSTRING, 0, (LPARAM) szBuffer) ;
			}
			if (nCurSel >= 0)
				SendMessage (hwndKeyToBind, CB_SETCURSEL, (WPARAM) nCurSel, (LPARAM) 0) ;
		}
		hwndJMap		= GetDlgItem (hDlg, IDC_COMBO_JMAPFUNCTOBIND) ;
		hwndJAbbrevMap	= GetDlgItem (hDlg, IDC_COMBO_JABBREVMAPFUNCTOBIND) ;
		if (hwndJMap != NULL && hwndJAbbrevMap != NULL) {
			for (i = 0 ; i < NUM_SELECTABLE_FUNCTIONS ; i ++) {
				SendMessage (hwndJMap, CB_ADDSTRING, 0, (LPARAM) srszKeyFuncName [i]) ;
				SendMessage (hwndJAbbrevMap, CB_ADDSTRING, 0, (LPARAM) srszKeyFuncName [i]) ;
			}
			if (nJMapFunc >= 0)
				SendMessage (hwndJMap, CB_SETCURSEL, (WPARAM) nJMapFunc, (LPARAM) 0) ;
			if (nJAbbrevFunc >= 0)
				SendMessage (hwndJAbbrevMap, CB_SETCURSEL, (WPARAM) nJAbbrevFunc, (LPARAM) 0) ;
		}
		return	TRUE ;
	}
	case	WM_COMMAND:
		/*	[ށBԂl global ϐoR邵Ȃ... */
		switch (LOWORD (wParam)) {
		case	IDOK:
		{
			register int	nKeyCodeInt, nJMapBind, nJAbbrevBind ;

			/*	ݑIĂ item 𓾂BCB_ERR ̏ꍇ undefined 
			 *	Ǝv...ȁB*/
			pArg	= (TDlgSelectKeyBindArg *)GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg != NULL) {
				hwndKeyToBind	= GetDlgItem (hDlg, IDC_COMBO_KEYTOBIND) ;
				hwndJMap		= GetDlgItem (hDlg, IDC_COMBO_JMAPFUNCTOBIND) ;
				hwndJAbbrevMap	= GetDlgItem (hDlg, IDC_COMBO_JABBREVMAPFUNCTOBIND) ;
				nKeyCodeInt		= SendMessage (hwndKeyToBind,  CB_GETCURSEL, 0, 0) ;
				nJMapBind		= SendMessage (hwndJMap,       CB_GETCURSEL, 0, 0) ;
				nJAbbrevBind	= SendMessage (hwndJAbbrevMap, CB_GETCURSEL, 0, 0) ;
				if (nKeyCodeInt 	!= CB_ERR &&
					nJMapBind   	!= CB_ERR &&
					nJAbbrevBind	!= CB_ERR) {
					DEBUGPRINTF ((MYTEXT ("Select (%d, %d, %d)\n"),
								  nKeyCodeInt, nJMapBind, nJAbbrevBind)) ;
					pArg->m_nCurSel			= nKeyCodeInt ;
					pArg->m_nJMapBind		= nJMapBind ;
					pArg->m_nJAbbrevBind	= nJAbbrevBind ;
				} else {
					/*	ƒlĂȂ CANCEL ƈB*/
					pArg->m_nCurSel			= -1 ;
					pArg->m_nJMapBind		= NFUNC_INVALID_CHAR ;
					pArg->m_nJAbbrevBind	= NFUNC_INVALID_CHAR ;
					wParam					= IDCANCEL ;
				}
			}
		}
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		}
	default:
		break ;
	}
	return	FALSE ;
}

int CALLBACK
skkimeConfig_compareKeyBind (
	register LPARAM		lParam1,
	register LPARAM		lParam2, 
	register LPARAM		lParamSort)
{
	return	(lParam1 - lParam2) ;
	UNREFERENCED_PARAMETER (lParamSort) ;
}

/*	L[oCh̓WXg BINARY ŏĂB
 */
BOOL
skkimeConfig_initKeymap (
	register LPCMYSTR		pstrSubKey,
	register BYTE*			pJMap,
	register BYTE*			pAbbrevMap)
{
	register BOOL	fFoundAbbrevMap, fFoundNormalMap ;
	register int	i ;
	HKEY	hSubKey ;
	BYTE	rbyKeyMap [SIZE_MYKEYMAP] ;

	fFoundNormalMap	= FALSE ;
	fFoundAbbrevMap	= FALSE ;

	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, pstrSubKey, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		DWORD	cbData, dwType ;

		DEBUGPRINTF ((MYTEXT ("RegKeyOpenEx succeeded (\"%s\")\n"), pstrSubKey)) ;

		/*	j-map ̏B*/
		cbData	= sizeof (rbyKeyMap) ;
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_JMAP, NULL, &dwType, rbyKeyMap, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY) {
			register int	nData	= (int) cbData ;
			/*	o^Ăꍇ̏B*/
			for (i = 0 ; i < nData ; i ++) {
				if (rbyKeyMap [i] >= NUM_SELECTABLE_FUNCTIONS) {
					*(pJMap + i)	= NFUNC_INVALID_CHAR ;
				} else {
					*(pJMap + i)	= rbyKeyMap [i] ;
				}
			}
			fFoundNormalMap	= TRUE ;
		}
		/*	j-abbrev-map ̏B*/
		cbData	= sizeof (rbyKeyMap) ;
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_JABBREVMAP, NULL, &dwType, rbyKeyMap, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY) {
			register int	nData	= (int) cbData ;
			/*	o^Ăꍇ̏B*/
			for (i = 0 ; i < nData ; i ++) {
				if (rbyKeyMap [i] >= NUM_SELECTABLE_FUNCTIONS) {
					*(pAbbrevMap + i)	= NFUNC_INVALID_CHAR ;
				} else {
					*(pAbbrevMap + i)	= rbyKeyMap [i] ;
				}
			}
			fFoundAbbrevMap	= TRUE ;
		}
		MyRegCloseKey (hSubKey) ;
	}
	if (! fFoundNormalMap) {
		for (i = 0 ; i < SIZE_MYKEYMAP ; i ++) 
			*(pJMap + i)	= (BYTE)riDefaultSkkJMapBind [i] ;
	}
	if (! fFoundAbbrevMap) {
		for (i = 0 ; i < SIZE_MYKEYMAP ; i ++)
			*(pAbbrevMap + i)	= (BYTE)riDefaultAbbrevMapBind [i] ;
	}
	return	TRUE ;
}

/*	XVʂWXgɏosB
 */
BOOL
skkimeConfig_updateKeymap (
	register LPCMYSTR	pstrSubKey,
	register BYTE*		pJMap,
	register BYTE*		pAbbrevMap)
{
	HKEY			hSubKey ;
	register BOOL	fRetval	= TRUE ;

	if (! skkimeConfig_CreateKey (pstrSubKey, FALSE, &hSubKey))
		return	FALSE ;
	
	if (MyRegSetValueEx (hSubKey, REGSUBKEY_JMAP,		0,
						 REG_BINARY, pJMap,      SIZE_MYKEYMAP) != ERROR_SUCCESS ||
		MyRegSetValueEx (hSubKey, REGSUBKEY_JABBREVMAP,	0,
						 REG_BINARY, pAbbrevMap, SIZE_MYKEYMAP) != ERROR_SUCCESS) {
		fRetval	= FALSE ;
	}
	if (RegCloseKey (hSubKey) != ERROR_SUCCESS)
		fRetval	= FALSE ;
	return	fRetval ;
}

int
skkimeConfig_key2item (
	register BYTE*		pKeymap,
	register UINT		uKey)
{
	if (uKey >= SIZE_MYKEYMAP)
		return	NFUNC_NOFUNCTION ;
	return	*(pKeymap + uKey) ;
}

