#if !defined (UNITTEST)
#include "globals.h"
#else
#include "StdAfx.h"
#endif
#include "RuleTreeNode.h"
#include "keymap.h"

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputStringPair::pCreateInstance (
	LPCWSTR							wstrLeft,
	int								nLeftLen,
	LPCWSTR							wstrRight,
	int								nRightLen)
{
	CSkkRuleTreeNodeOutputStringPair*	pNode ;
	WCHAR*	ptr ;

	pNode	= new CSkkRuleTreeNodeOutputStringPair () ;
	if (pNode == NULL)
		return	NULL ;
	if (nLeftLen > 0 && wstrLeft != NULL) {
		ptr	= new WCHAR [nLeftLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, wstrLeft, nLeftLen * sizeof (WCHAR)) ;
		pNode->m_strHira	= ptr ;
		pNode->m_nHira		= nLeftLen ;
	}
	if (nRightLen > 0 && wstrRight != NULL) {
		ptr	= new WCHAR [nRightLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, wstrRight, nRightLen * sizeof (WCHAR)) ;
		pNode->m_strKata	= ptr ;
		pNode->m_nKata		= nRightLen ;
	}
	return	pNode ;
}

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputString::pCreateInstance (
	LPCWSTR							wstrString,
	int								nStringLen)
{
	CSkkRuleTreeNodeOutputString*	pNode ;
	WCHAR*	ptr ;

	pNode	= new CSkkRuleTreeNodeOutputString () ;
	if (pNode == NULL)
		return	NULL ;
	if (wstrString != NULL && nStringLen > 0) {
		ptr	= new WCHAR [nStringLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, wstrString, nStringLen * sizeof (WCHAR)) ;
		pNode->m_strValue	= ptr ;
		pNode->m_nValue		= nStringLen ;
	} 
	return	pNode ;
}

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputMacro::pCreateInstance (int nFuncNo) 
{
	CSkkRuleTreeNodeOutputMacro*	pNode ;

	pNode	= new CSkkRuleTreeNodeOutputMacro () ;
	if (pNode == NULL)
		return	NULL ;
	pNode->m_nMacro	= nFuncNo ;
	return	pNode ;
}

/*===============================================================
 *	static public methods
 */
LPCWSTR
CSkkRuleTreeNode::pGetPrefix (
	const CSkkRuleTreeNode*			pNode,
	int*							pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->m_nPrefix ;
	return	pNode->m_pPrefix ;
}

LPCWSTR
CSkkRuleTreeNode::pGetPrefix (
	CSkkRuleTreeNode*				pNode,
	int*							pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->m_nPrefix ;
	return	pNode->m_pPrefix ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pCreateInstance (
	int								nChar, 
	LPCWSTR							wstrPrefix,
	int								nPrefixLen,
	LPCWSTR							wstrNext,
	int								nNextLen,
	CSkkRuleTreeNodeOutput*			pOutput)
{
	CSkkRuleTreeNode*	pNode ;
	WCHAR*				pwDest ;

	pNode	= new CSkkRuleTreeNode () ;
	if (pNode == NULL)
		return	NULL ;

	pNode->m_nCH	= nChar ;
	if (nPrefixLen > 0 && wstrPrefix != NULL) {
		pwDest		= new WCHAR [nPrefixLen] ;
		if (pwDest == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		pNode->m_pPrefix	= pwDest ;
		memcpy (pNode->m_pPrefix, wstrPrefix, sizeof (WCHAR) * nPrefixLen) ;
	} else {
		nPrefixLen		= 0 ;
	}
	pNode->m_nPrefix	= nPrefixLen ;

	if (nNextLen > 0 && wstrNext != NULL) {
		pwDest		= new WCHAR [nNextLen] ;
		pNode->m_pNextState	= pwDest ;
		memcpy (pNode->m_pNextState, wstrNext, sizeof (WCHAR) * nNextLen) ;
	} else {
		nNextLen	= 0 ;
	}
	pNode->m_nNextState	= nNextLen ;
	pNode->m_pOutput	= pOutput ;
	return	pNode ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pGetBranchList (
	CSkkRuleTreeNode*		pRoot)
{
	return	pRoot != NULL? pRoot->pGetChild () : NULL ;
}

const CSkkRuleTreeNode*
CSkkRuleTreeNode::pGetBranchList (
	const CSkkRuleTreeNode*		pRoot)
{
	return	pRoot != NULL? pRoot->pGetChild () : NULL ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pSelectBranch (
	CSkkRuleTreeNode*		pRoot,
	int						nChar)
{
	CSkkRuleTreeNode*	pNode ;

	if (pRoot == NULL)
		return	NULL ;

	pNode	= pRoot->pGetChild () ;
	while (pNode != NULL) {
		if (pNode->iGetChar () == nChar)
			break ;
		pNode	= pNode->pGetBrother () ;
	}
	return	pNode ;
}

const CSkkRuleTreeNode*
CSkkRuleTreeNode::pSelectBranch (
	const CSkkRuleTreeNode*		pRoot,
	int							nChar)
{
	const CSkkRuleTreeNode*	pNode ;

	if (pRoot == NULL)
		return	NULL ;

	pNode	= pRoot->pGetChild () ;
	while (pNode != NULL) {
		if (pNode->iGetChar () == nChar)
			break ;
		pNode	= pNode->pGetBrother () ;
	}
	return	pNode ;
}

/*	ԂĺATɐ(ǂȂȂ炻܂)
 *	̎_ł̃m[h ppResult ɓB
 */
int
CSkkRuleTreeNode::iSearchTree (
	CSkkRuleTreeNode*			pTree,
	LPCWSTR						pString,
	int							nString,
	CSkkRuleTreeNode**			ppResult)
{
	CSkkRuleTreeNode*	pParent ;
	CSkkRuleTreeNode*	pNode ;
	LPCWSTR		wptr ;
	int			nptr, nCH ;

	pNode		= pTree != NULL? pTree->pGetChild () : NULL ;
	pParent		= pTree ;
	wptr		= pString ;
	nptr		= nString ;
	while (nptr > 0) {
		nCH		= *wptr ;
		/* Z푤̒TB*/
		while (pNode != NULL) {
			if (pNode->iGetChar () == nCH)
				break ;
			pNode	= pNode->pGetBrother () ;
		}
		if (pNode == NULL) {
			/* łBvĂƂ܂ŕԂB*/
			if (ppResult != NULL)
				*ppResult	= pParent ;
#if defined (DEBUG)
			DEBUGPRINTF ((TEXT ("iSeachTree: not found\n"))) ;
#endif
			return	wptr - pString ;
		}
		wptr	++ ;
		nptr	-- ;
		if (nptr <= 0) {
			/* B*/
			if (ppResult != NULL)
				*ppResult	= pNode ;
#if defined (DEBUG)
			DEBUGPRINTF ((TEXT ("iSeachTree: found\n"))) ;
			pNode->vDebugOut () ;
			pParent->vDebugOut () ;
#endif
			return	wptr - pString ;
		}
#if defined (DEBUG)
		DEBUGPRINTF ((TEXT ("iSearchTree: -> goto child\n"))) ;
		pNode->vDebugOut () ;
#endif
		pParent	= pNode ;
		pNode	= pNode->pGetChild () ;
	}
	if (ppResult != NULL)
		*ppResult	= pTree ;
	return	0 ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	LPCWSTR				wstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	LPCWSTR				wstrHira,
	int					nHiraLen,
	LPCWSTR				wstrKata,
	int					nKataLen)
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint ;
	CSkkRuleTreeNode*		pAddTree ;
	int						nOffset = 0 ;
	CSkkRuleTreeNodeOutput*	pOutput	= NULL ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
			pAddPoint->vSetChild (pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= CSkkRuleTreeNodeOutputStringPair::pCreateInstance (wstrHira, nHiraLen, wstrKata, nKataLen) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			delete	pOutput ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTree) ;
	} else {
		pOutput		= CSkkRuleTreeNodeOutputStringPair::pCreateInstance (wstrHira, nHiraLen, wstrKata, nKataLen) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddPoint->vSetOutput (pOutput) ;
	}
	return	TRUE ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	LPCWSTR				wstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	LPCWSTR				wstrString,
	int					nString)
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint ;
	CSkkRuleTreeNode*		pAddTree ;
	CSkkRuleTreeNodeOutput*	pOutput	= NULL ;
	int						nOffset = 0 ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= CSkkRuleTreeNode::pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= CSkkRuleTreeNode::iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
			pAddPoint->vSetChild (pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= CSkkRuleTreeNodeOutputString::pCreateInstance (wstrString, nString) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			delete	pOutput ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTree) ;
	} else {
		pOutput		= CSkkRuleTreeNodeOutputString::pCreateInstance (wstrString, nString) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddPoint->vSetOutput (pOutput) ;
	}
	return	TRUE ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	LPCWSTR				wstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	int					nMacro) 
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint ;
	CSkkRuleTreeNode*		pAddTree ;
	int						nOffset = 0 ;
	CSkkRuleTreeNodeOutput*	pOutput	= NULL ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= CSkkRuleTreeNode::pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= CSkkRuleTreeNode::iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
			pAddPoint->vSetChild (pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= CSkkRuleTreeNodeOutputMacro::pCreateInstance (nMacro) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= CSkkRuleTreeNode::pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			delete	pOutput ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		pAddTree->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTree) ;
	} else {
		pOutput		= CSkkRuleTreeNodeOutputMacro::pCreateInstance (nMacro) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddPoint->vSetOutput (pOutput) ;
	}
	return	TRUE ;
}

BOOL
CSkkRuleTreeNode::bDeleteRule (
	CSkkRuleTreeNode**	ppTree,
	LPCWSTR				pString,
	int					nString)
{
	CSkkRuleTreeNode*	pTree ;
	CSkkRuleTreeNode*	pParent ;
	CSkkRuleTreeNode*	pNode ;
	CSkkRuleTreeNode*	pPrev ;
	LPCWSTR		wptr ;
	int			nptr, nCH ;

	pTree	= *ppTree ;
	pNode	= pTree != NULL? pTree->pGetChild () : NULL ;
	pParent	= pTree ;
	wptr	= pString ;
	nptr	= nString ;

	/*	܂ brother ւ̒TsB*/
	while (pNode != NULL && nptr > 0) {
		nCH		= *wptr ;
		pPrev	= NULL ;
		while (pNode != NULL) {
			if (pNode->iGetChar () == nCH) 
				break ;
			pPrev	= pNode ;
			pNode	= pNode->pGetBrother () ;
		}
		/* YȂB*/
		if (pNode == NULL)
			return	FALSE ;

		wptr	++ ;
		nptr	-- ;
		if (nptr <= 0) {
			/* ̈ʒuō폜B*/
			if (pPrev == NULL) {
				/* ẽN\ւB*/
				if (pParent != NULL) {
					pParent->vSetChild (pNode->pGetBrother ()) ;
				} else {
					*ppTree	= pNode->pGetBrother () ;
				}
				pNode->vSetBrother (NULL) ;
			} else {
				/*	Z̃N\ւB
				 *	cqq͑SłB
				 */
				pPrev->vSetBrother (pNode->pGetBrother ()) ;
				pNode->vSetBrother (NULL) ;
			}
			CSkkRuleTreeNode::vDestroyTree (pNode) ;
			break ;
		}
		pParent	= pNode ;
		pNode	= pNode->pGetChild () ;
	}
	return	TRUE ;
}

void
CSkkRuleTreeNode::vDestroyTree (
	CSkkRuleTreeNode*	pRoot)
{
	CSkkRuleTreeNode*	pBrother ;
	CSkkRuleTreeNode*	pChild ;

	if (pRoot == NULL)
		return ;

	pBrother	= pRoot->pGetBrother () ;
	pChild		= pRoot->pGetChild () ;
	pRoot->vSetBrother (NULL) ;
	pRoot->vSetChild (NULL) ;
	delete	pRoot ;

	if (pBrother != NULL)
		CSkkRuleTreeNode::vDestroyTree (pBrother) ;
	if (pChild != NULL)
		CSkkRuleTreeNode::vDestroyTree (pChild) ;
	return ;
}

/*===============================================================
 *	public methods
 */
void
CSkkRuleTreeNode::vDebugOut () const 
{
	WCHAR	buf [256] ;
	WCHAR	bufPrefix [128], bufNextState [128] ;
	int		n ;

	if (m_pPrefix != NULL && m_nPrefix > 0) {
		n	= (m_nPrefix < (MYARRAYSIZE (bufPrefix) - 1))? m_nPrefix : (MYARRAYSIZE (bufPrefix) - 1) ;
		memcpy (bufPrefix, m_pPrefix, n * sizeof (WCHAR)) ;
		bufPrefix [n]	= L'\0' ;
	} else {
		bufPrefix [0]	= L'\0' ;
	}
	if (m_pNextState != NULL && m_nNextState > 0) {
		n	= (m_nNextState < (MYARRAYSIZE (bufNextState) - 1))? m_nNextState : (MYARRAYSIZE (bufNextState) - 1) ;
		memcpy (bufNextState, m_pNextState, n * sizeof (WCHAR)) ;
		bufNextState [n]	= L'\0' ;
	} else {
		bufNextState [0]	= L'\0' ;
	}
	if (m_pOutput != NULL) {
		switch (m_pOutput->iGetType ()) {
		case	CSkkRuleTreeNodeOutput::NTYPE_STRINGPAIR:
			{
				CSkkRuleTreeNodeOutputStringPair*	pOutput	= static_cast <CSkkRuleTreeNodeOutputStringPair*> (m_pOutput) ;
				LPCWSTR	pHira, pKata ;
				int		nHira, nKata ;
				WCHAR	bufHira [128], bufKata [128] ;

				pHira	= pOutput->pGetHiragana (&nHira) ;
				pKata	= pOutput->pGetKatakana (&nKata) ;
				if (pHira != NULL && nHira > 0) {
					n	= (nHira < (MYARRAYSIZE (bufHira) - 1))? nHira : (MYARRAYSIZE (bufHira) - 1) ;
					memcpy (bufHira, pHira, nHira * sizeof (WCHAR)) ;
					bufHira [n]	= L'\0' ;
				} else {
					bufHira [0]	= L'\0' ;
				}
				if (pKata != NULL && nKata > 0) {
					n	= (nKata < (MYARRAYSIZE (bufKata) - 1))? nKata : (MYARRAYSIZE (bufKata) - 1) ;
					memcpy (bufKata, pKata, nKata * sizeof (WCHAR)) ;
					bufKata [n]	= L'\0' ;
				} else {
					bufKata [0]	= L'\0' ;
				}
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Hira(\"%s\"), Kata(\"%s\")\n", bufPrefix, bufNextState, bufHira, bufKata) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	CSkkRuleTreeNodeOutput::NTYPE_STRING:
			{
				CSkkRuleTreeNodeOutputString*	pOutput	= static_cast <CSkkRuleTreeNodeOutputString*> (m_pOutput) ;
				WCHAR	bufString [128] ;
				LPCWSTR	pString ;
				int		nString ;

				pString	= pOutput->pGetString (&nString) ;
				if (pString != NULL && nString > 0) {
					n	= (nString < (MYARRAYSIZE (bufString) - 1))? nString : (MYARRAYSIZE (bufString) - 1) ;
					memcpy (bufString, pString, nString * sizeof (WCHAR)) ;
				} else {
					bufString [0]	= L'\0' ;
				}
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), String(\"%s\")\n", bufPrefix, bufNextState, bufString) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	CSkkRuleTreeNodeOutput::NTYPE_MACRO:
			{
				CSkkRuleTreeNodeOutputMacro*	pOutput	= static_cast <CSkkRuleTreeNodeOutputMacro*> (m_pOutput) ;
				int		nMacro ;

				nMacro	= pOutput->iGetFunctionCode () ;
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Macro(%d)\n", bufPrefix, bufNextState, nMacro) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;
		default:
			break ;
		}
		wprintf (L"%s\n", buf) ;
	} else {
		wprintf (L"nil\n", buf) ;
	}
	return ;
}

#if 0 //defined (UNIT_TEST)
/*	P̃eXĝ߂̃R[hB
 */
int
_tmain (void)
{
	CSkkRuleTreeNode*	pRoot	= NULL ;
	CSkkRuleTreeNode*	pNode	= NULL ;
	int		i, n ;
	char	buf [12] ;

	/*
struct TSkkBaseRuleT1 {
	LPCTSTR	_strState ;
	LPCTSTR	_strNext ;
	struct {
		LPCTSTR	_strHira ;
		LPCTSTR	_strKata ;
	}	_case ;
} ;*/
	_tsetlocale (LC_CTYPE, TEXT ("")) ;

	/* debugger  process  attach 邽߂̗]TB*/
	_tprintf (TEXT ("> ")) ;
	fflush (stdout) ;
	fgets (buf, 10, stdin) ;

	for (i = 0 ; i < MYARRAYSIZE (_rT1) ; i ++) {
		int	nPrefix, nNextState, nHira, nKata ;

		nPrefix		= (_rT1 [i]._strState      != NULL)? lstrlenW (_rT1 [i]._strState)      : 0 ;
		nNextState	= (_rT1 [i]._strNext       != NULL)? lstrlenW (_rT1 [i]._strNext)       : 0 ;
		nHira		= (_rT1 [i]._case._strHira != NULL)? lstrlenW (_rT1 [i]._case._strHira) : 0 ;
		nKata		= (_rT1 [i]._case._strKata != NULL)? lstrlenW (_rT1 [i]._case._strKata) : 0 ;
		if (! TSkkRuleTreeNode_bAddRule (&pRoot, _rT1 [i]._strState, nPrefix, _rT1 [i]._strNext, nNextState, _rT1 [i]._case._strHira, nHira, _rT1 [i]._case._strKata, nKata)) {
			_tprintf (TEXT ("TSkkRuleTreeNode_bAddRule: failed. (%d)\n"), i) ;
			break ;
		}
	}
	if (pRoot != NULL) {
		_tprintf (TEXT ("Root ...\n")) ;
		TSkkRuleTreeNode_vDebugOut (pRoot) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"b", 1, &pNode) ;
	_tprintf (TEXT ("Search ... \"b\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"fya", 3, &pNode) ;
	_tprintf (TEXT ("Search ... \"fya\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	return	0 ;
}
#endif



